#include <NDS/nds.h>
#include <ds_misc.h>
#include "c_defs.h"

#include "m_nsf.h"
#include "audiosys.h"
#include "handler.h"

#define MIXFREQ 0x5e00
#define MIXBUFSIZE 64

s16 buffer[MIXBUFSIZE*2];
//s16 *buffer=(s16*)0x27f9000;

u8 SPIwriteread(u8 data);
void touch_update(void);
void readAPU(void);
void resetAPU(void);

void restartsound(int chan) { 
	TIMER1_CR=0;

	SCHANNEL_CR(chan)=SCHANNEL_ENABLE|SOUND_ONE_SHOT|SOUND_VOL(0x7F)|SOUND_PAN(0x40)|SOUND_16BIT;

	TIMER0_DATA = (u16)TIMER_FREQ(MIXFREQ);
	TIMER0_CR = TIMER_ENABLE; 
	TIMER1_DATA = 0x10000 - MIXBUFSIZE;
	TIMER1_CR = TIMER_CASCADE | TIMER_IRQ_REQ | TIMER_ENABLE;
}

int APU_paused=0;
void mix(int chan) {
	if(!APU_paused) {
		NESAudioRender(&buffer[chan*MIXBUFSIZE],MIXBUFSIZE);
		readAPU();
	}
}

void initsound() { 		
	powerON(POWER_SOUND); 
	SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);
	SCHANNEL_SOURCE(0)=(u32)&buffer[0];
	SCHANNEL_SOURCE(1)=(u32)&buffer[MIXBUFSIZE];
	SCHANNEL_TIMER(0)=SOUND_FREQ(MIXFREQ); 
	SCHANNEL_TIMER(1)=SOUND_FREQ(MIXFREQ); 
	SCHANNEL_LENGTH(0)=MIXBUFSIZE/2;
	SCHANNEL_LENGTH(1)=MIXBUFSIZE/2;
	SCHANNEL_REPEAT_POINT(0) = 0; 
	SCHANNEL_REPEAT_POINT(1) = 0; 
} 

int soundinterrupt() {
	static int chan=0;
	restartsound(chan);
	chan^=1;
	mix(chan);
	return IRQ_TIMER1;
}

int oldPM;
void bios_SPI_write(int device,int data,int select);
void writePM(int flags) {
	oldPM=flags;
	bios_SPI_write(SPI_POWER,0,1);
	bios_SPI_write(SPI_POWER,flags,0);
}

u32 fifo_recv() {
	u16 cnt=REG_FIFO_CNT;
	if (cnt&FIFO_CNT_ERR) {
		REG_FIFO_CNT=cnt|FIFO_CNT_SEND_CLEAR;
		return 0;
	} else if(REG_FIFO_CNT & FIFO_CNT_RECV_EMPTY)
		return 0;
	return REG_FIFO_RECV;
}

void APUSoundWrite(Uint address, Uint value);	//from s_apu.c (skip using read handlers, just write it directly)
int fifointerrupt() {
	u32 msg;

	do {
		msg=fifo_recv();
		switch(msg&0xff) {
			case FIFO_WRITEPM:
				writePM(msg>>8);
				break;
			case FIFO_APU_PAUSE:
				APU_paused=1;
				memset(buffer,0,sizeof(buffer));
				//swiHalt();
				break;
			case FIFO_UNPAUSE:
				APU_paused=0;
				break;
//			case FIFO_SETVOLUME:
//				NESVolume(msg>>8);
//				break;
			case FIFO_APU_RESET:
				APU_paused=0;
				resetAPU();
				break;
		}
	} while(msg);
	return IRQ_FIFO_FULL;
}

int powerinterrupt() {
	return IRQ_PM;
}

int vblankinterrupt() {
	return IRQ_VBLANK;
}

void interrupthandler() {
	u32 flags=IF&IE;
	if(flags&IRQ_TIMER1)
		soundinterrupt();
	if(flags&IRQ_FIFO_FULL)
		fifointerrupt();
	if(flags&IRQ_PM)
		powerinterrupt();
//	if(flags&IRQ_VBLANK)
//		vblankinterrupt();
	VBLANK_INTR_WAIT_FLAGS|=flags;
	IF=flags;
}

u8 SPIwriteread(u8 data) {
	while (SERIAL_CR & SERIAL_BUSY);
	SERIAL_DATA = data;
	while (SERIAL_CR & SERIAL_BUSY);
	return SERIAL_DATA;
}


u16 readtouch(u32 command) {
	u16 result;

	SERIAL_CR = SERIAL_ENABLE | 0xA01;
	SERIAL_DATA = command;
	while (SERIAL_CR & SERIAL_BUSY);

	SERIAL_DATA = 0;
	while (SERIAL_CR & SERIAL_BUSY);
	result = SERIAL_DATA;

	SERIAL_CR = SERIAL_ENABLE | 0x201;
	SERIAL_DATA = 0;
	while (SERIAL_CR & SERIAL_BUSY);

	return ((result & 0x7F) << 5) | (SERIAL_DATA >> 3);
}

void touch_update() {
	int cx1=TOUCH_CAL_X1;
	int cx2=TOUCH_CAL_X2;
	int cy1=TOUCH_CAL_Y1;
	int cy2=TOUCH_CAL_Y2;
	int nx1=TOUCH_CNTRL_X1;
	int nx2=TOUCH_CNTRL_X2;
	int ny1=TOUCH_CNTRL_Y1;
	int ny2=TOUCH_CNTRL_Y2;
	int tx=readtouch(TSC_MEASURE_X);
	int ty=readtouch(TSC_MEASURE_Y);

	int x = (tx-cx1) * (nx2-nx1) / (cx2-cx1) + nx1;
	int y = (ty-cy1) * (ny2-ny1) / (cy2-cy1) + ny1;

	if(x<0) x=0;
	if(x>255) x=255;
	if(y<0) y=0;
	if(y>191) y=191;

	IPC_TOUCH_X=x;
	IPC_TOUCH_Y=y;
}

void DS_init() {
	REG_FIFO_CNT=FIFO_CNT_ENABLE | FIFO_CNT_SEND_CLEAR | FIFO_CNT_RECV_IRQ | FIFO_CNT_ERR;
	DISP_SR=DISP_VBLANK_IRQ;
	IRQ_HANDLER=interrupthandler;
	IE=IRQ_VBLANK|IRQ_TIMER1|IRQ_FIFO_FULL;
	IF=~0;
	VBLANK_INTR_WAIT_FLAGS=0;
	IME=IME_ENABLED;
}

void NES6502Irq(void) {}	//trigger 6502 IRQ
Uint NES6502ReadDma(Uint A) {return 0;}

void resetAPU() {
	IPC_APU_WRITE_HEAD=0;
	IPC_APU_WRITE_TAIL=0;
	NESReset();
}

void readAPU() {
	u32 *buff=soundbuff;
	int head=IPC_APU_WRITE_HEAD;
	int tail=IPC_APU_WRITE_TAIL;
	u32 val;
	
	while(tail!=head) {
		val=buff[tail];
		APUSoundWrite(val&0xffff,(val>>16)&0xff);
		tail++;
		tail&=255;
	}
	IPC_APU_WRITE_TAIL=tail;
}

void ARM7sleep() {
	u32 ie_save, pm_save;
	pm_save=oldPM;
	writePM(oldPM|PM_LED_SLOW);
	swiChangeSoundBias(0,0x400);
	ie_save=IE;
	IE=IRQ_PM;
	swiHalt();
	IF=~0;
	IE=ie_save;
	writePM(pm_save);
	swiChangeSoundBias(1,0x400);
	IPC_SLEEP=0;
}

void arm7main() {

	DS_init();


	NESAudioFrequencySet(MIXFREQ);
	//NESAudioFilterSet(NES_AUDIO_FILTER_NONE);
	NSFLoad();
	resetAPU();
	NESVolume(0);
	
	initsound();
	restartsound(0);

	while(1) {
		swiWaitForVBlank();
		touch_update();
		IPC_KEYS=(KEYS|(XKEYS<<10))^0x1ffff;
		IPC_ALIVE++;
		if(IPC_SLEEP)
			ARM7sleep();
	}
}
