#include <NDS/nds.h>
#include <ds_misc.h>

extern u16 font;
extern u16 fontpal;

static int getkb() {
	static int oldkey=0;
	int k,x;
	k=READ_KEYS;
	x=(oldkey^k)&k;
	oldkey=k;
	return x;
}

void __rt_div0() {
}

static inline void dmaFixed(const void * source, void * dest, uint32 size)
{
	DMA_SRC(3) = (uint32)source;
	DMA_DEST(3) = (uint32)dest;
	DMA_CR(3) = DMA_SRC_FIX | DMA_COPY_HALFWORDS | (size>>1);
	while(DMA_CR(3) & DMA_BUSY);
}

static void gfx_init() {
	u16 d;
	POWER_CR=POWER_ALL_2D|POWER_SWAP_LCDS;

	vramSetBanks(VRAM_A_MAIN_BG,VRAM_B_MAIN_SPRITE,VRAM_C_SUB_BG,VRAM_D_SUB_SPRITE);

	DISPLAY_CR=MODE_0_2D|DISPLAY_BG0_ACTIVE;
	BG0_CR=BG_TILE_BASE(1);
	BRIGHTNESS=0;
	BG0_X0=0;
	BG0_Y0=0;
	BLEND_CR=0;
	SUB_DISPLAY_CR=MODE_0_2D|DISPLAY_BG0_ACTIVE;
	SUB_BG0_CR=BG_TILE_BASE(1);
	SUB_BRIGHTNESS=0;
	SUB_BG0_X0=0;
	SUB_BG0_Y0=0;
	SUB_BLEND_CR=0;

	dmaCopyHalfWords(3,&font,(u16*)(CHAR_BASE_BLOCK(1)+32*32),1024*3);
	dmaCopyHalfWords(3,&fontpal,(u16*)BG_PALETTE,64);
	dmaCopyHalfWords(3,&font,(u16*)(CHAR_BASE_BLOCK_SUB(1)+32*32),1024*3);
	dmaCopyHalfWords(3,&fontpal,(u16*)BG_PALETTE_SUB,64);
	d=' ';
	dmaFixed(&d,(u16*)SCREEN_BASE_BLOCK(0),64*24);
	dmaFixed(&d,(u16*)SCREEN_BASE_BLOCK_SUB(0),64*24);
}

static void text(int offset,int screen,char *s) {
	u16 *p;
	if(!screen)
		p=(u16*)SCREEN_BASE_BLOCK(0);
	else
		p=(u16*)SCREEN_BASE_BLOCK_SUB(0);
	p+=offset;
	while(*s) {
		*p++=*s++;
	}
}

static void hex32(int offset,int screen,int d,int hi) {
	u16 *p;
	u16 c;
	int i;
	
	if(!screen)
		p=(u16*)SCREEN_BASE_BLOCK(0);
	else
		p=(u16*)SCREEN_BASE_BLOCK_SUB(0);
	p+=offset;
	for(i=7;i>=0;i--) {
		c=d&0x0f;
		if(c<10) c+='0';
		else c=c-10+'A';
		if(!hi) c|=0x1000;
		d>>=4;
		p[i]=c;
		hi--;
	}
}

static void dumpmem(u8 *addr) {
	u16 *p;
	u16 c;
	int i,j;
	u32 d;
	int hi=0;
	
	p=(u16*)(SCREEN_BASE_BLOCK_SUB(0)+64);
	for(j=0;j<23*16;j++) {
		d=addr[j];
		if(!(j&15)) hi++;
		for(i=1;i>=0;i--) {
			c=d&0x0f;
			if(c<10) c+='0';
			else c=c-10+'A';
			if(hi&1) c|=0x1000;
			d>>=4;
			p[i]=c;
		}
		p++;
		p++;
		hi++;
	}
}

static void vblank() {
	while(DISP_SR&1);
	while(!DISP_SR&1);
}

static inline u32 getPID() {
	u32 ret;
	__asm {mrc p15,0,r0,c13,c0,1; mov ret,r0}
	return ret;
}

//regs: sp,lr,r0-r11,xx,xx,spsr,r12,pc
char *rtxt[]={" r0"," r1"," r2"," r3"," r4"," r5"," r6"," r7"," r8"," r9","r10","r11","r12"," sp"," lr"," pc","psr","pid"};
void exception(u32 *regs,u32 cpsr) {
	int i;
	int k;
	int sel=0;
	int addr;

	DMA0_CR=0;
	DMA1_CR=0;
	DMA2_CR=0;
	DMA3_CR=0;
	
	gfx_init();

	text(0,0,"Crash!");
	cpsr&=0x1f;
	text(7+0*32,0,cpsr==0x17?"(Abort)":"(Undefined)");
	for(i=0;i<18;i++) {
		text(64+i*16,0,rtxt[i]);
		hex32(68+i*16,0,regs[i+2],8);
	}
	hex32(68+12*16,0,regs[17],8);			//r12
	hex32(68+13*16,0,regs[0],8);			//sp
	hex32(68+14*16,0,regs[1],8);			//lr
	hex32(68+15*16,0,regs[18]-(cpsr==0x17?8:4),8);	//pc
	hex32(68+16*16,0,regs[16],8);			//cpsr
	hex32(68+17*16,0,getPID(),8);

	addr=regs[0];
	k=KEY_UP;
	while(1) {
		vblank();
		if(k) {
			switch(k) {
				case KEY_A:
					i=sel?1:4;
					addr+=i<<((sel&7)*4);
					break;
				case KEY_B:
					i=sel?1:4;
					addr-=i<<((sel&7)*4);
					break;
				case KEY_LEFT:
					sel++;
					if(sel>=8) sel--;
					break;
				case KEY_RIGHT:
					sel--;
					if(sel<0) sel=0;
					break;
			}
		}
		hex32(0,1,addr,sel);
		dumpmem((u8*)addr);
		k=getkb();
	}
}
