#include <nds/nds.h>
#include <ds_misc.h>
#include <string.h>
#include "c_defs.h"
#include "dldi/fat.h"
#include "dldi/dldi.h"

int freemem_start;	//points to unused mem (end of loaded rom)
u32 oldinput;

char romfilename[256];	//name of current rom.  null when there's no file (rom was embedded)
char *romfileext;	//points to file extension

//extern u8 autostate;			//from ui.c

void rommenu(int roms);
void drawmenu(int sel,int roms);
int getinput(void);

int init_rommenu(void);
void listrom(int line,int rom,int highlight); //print rom title
int loadrom(int rom);	//return -1 on success, or rom count (from directory change)
void stringsort(char**);

void do_rommenu() {
	int roms=init_rommenu();
	if(!roms) {
		if(!(DLDI_interface.ul_Features & (FEATURE_SLOT_GBA|FEATURE_SLOT_NDS))) {	//still have the default DLDI driver?
			consoletext(64*3,"DLDI driver is missing.",0);
			consoletext(64*4,"http://dldi.drunkencoders.com/",0);
		} else if(!active_interface) {			//another driver is present, but init failed
			consoletext(64*3,DLDI_string,0);
			consoletext(64*4,"Device failed.",0);
		} else						//no DLDI error, no files
			consoletext(64*3,"No roms found.",0);
		while(1) swiSleep();
	}
	rommenu(roms);
	clearconsole();
}

void rommenu(int roms) {
	static int selectedrom=0;
	int key;
	int sel=selectedrom;
	int loaded=0;

	//save_sram();

	oldinput=IPC_KEYS;

	if(roms==1)
		key=KEY_START;
	else
		key=0x80000000;
	do {
		swiWaitForVBlank();
		switch(key) {
			case KEY_START:
			case KEY_B:
			case KEY_A:
			case KEY_Y:
				loaded=loadrom(sel);
				if(loaded>=0) {	//didn't load (it was a directory)
					sel=0;
					roms=loaded;
				}
				break;
			case KEY_RIGHT:
				sel+=10;
				if(sel>roms-1) sel=roms-1;
				break;
			case KEY_LEFT:
				sel-=10;
				if(sel<0) sel=0;
				break;
			case KEY_UP:
				sel=sel+roms-1;
				break;
			case KEY_DOWN:
				sel++;
				break;
		}
		sel%=roms;
		if(key && loaded>=0)
			drawmenu(sel,roms);
		key=getinput();
	} while(loaded>=0);

	selectedrom=sel;

	load_sram();

	//if(autostate) loadstate(..);
}

#define ROWS 21
#define OFFSET 3
void drawmenu(int sel,int roms) {
	int i,j,topline,rom;
	
	clearconsole();
	if(roms>ROWS) {
		topline=8*(roms-ROWS)*sel/(roms-1);
		rom=topline/8;
		j=(rom<roms-ROWS)?ROWS+1:ROWS;
	} else {
		topline=0;
		rom=0;
		j=roms;
	}
	
	for(i=0;i<j;i++) {
		listrom(i+OFFSET,rom,sel==rom);
		rom++;
	}

//	if(roms>ROWS)
		SUB_BG0_Y0=topline%8;
//	else
//		SUB_BG0_Y0=-(ROWS*4)+roms*4;
}

int getinput() {
	static int lastdpad,repeatcount=0;
	int dpad;
	int keyhit=(oldinput^IPC_KEYS)&IPC_KEYS;
	oldinput=IPC_KEYS;

	dpad=IPC_KEYS&(KEY_UP+KEY_DOWN+KEY_LEFT+KEY_RIGHT);
	if(lastdpad==dpad) {
		repeatcount++;
		if(repeatcount<25 || repeatcount&3)	//delay/repeat
			dpad=0;
	} else {
		repeatcount=0;
		lastdpad=dpad;
	}
	return dpad|(keyhit&(KEY_Y+KEY_X+KEY_A+KEY_B+KEY_START));
}

void listrom(int line,int rom,int highlight) {
	char *s;
	char **files=(char**)&Image$$EWRAM$$ZI$$Limit;
	s=files[rom];
	if(*s==1) {	//dir
		menutext(line,s+1,highlight?2:0);
	} else		//file
		menutext(line,s+1,highlight);
}

int loadrom(int rom) {
	int i;
	char **files=(char**)&Image$$EWRAM$$ZI$$Limit;
	FAT_FILE *f;
	
	if(*files[rom]==1) {	//directory
		FAT_chdir(files[rom]+1);
		return init_rommenu();
	} else {	//file
		memcpy(romfilename,files[rom]+1,256);
		romfileext=strrchr(romfilename,'.')+1;
		f=FAT_fopen(files[rom]+1,"r");
		i=FAT_fread(&Image$$EWRAM$$ZI$$Limit,1,0x200000,f);	//read the whole thing (filesize=some huge number, don't care)
		freemem_start=((u32)&Image$$EWRAM$$ZI$$Limit+i+3)&~3;
		FAT_fclose(f);
		initcart((u8*)&Image$$EWRAM$$ZI$$Limit);
		return -1;	//(-1 on success)
	}
}

#define MAXFILES 0x4000
//return number of file entries (roms+dirs)
int init_rommenu() {
	char **files=(char**)&Image$$EWRAM$$ZI$$Limit;
	char *nextfile=(char*)&files[MAXFILES];
	int i,idx=0;

	if(!active_interface)		//DLDI trouble
		return 0;

	i=FAT_FindFirstFile(nextfile+1);
	while(i && (idx<MAXFILES-1)) {
		if(strstr(nextfile+1,".NES") || (i==2 && (nextfile[2]!=0 || nextfile[1]!='.'))) {
			*nextfile=3-i;		//type:  1=dir, 2=file
			FAT_GetLongFilename(nextfile+1);
			files[idx]=nextfile;
			nextfile+=strlen(nextfile)+1;
			idx++;
		}
		i=FAT_FindNextFile(nextfile+1);
	}
	files[idx]=0;
	stringsort(files);
	return idx;
}

//case insensitive string compare
inline int less(char *s1,char *s2) {
	int c1,c2;
	do {
		c1=*s1++;
		c2=*s2++;
		if(c1>='a') c1=c1-'a'+'A';
		if(c2>='a') c2=c2-'a'+'A';
	} while(c1==c2);
	return c1<c2;
}

//lazy man's string sorter
void stringsort(char **p1) {
	char **p2;
	char *s1,*s2;
	do {	//next s1
		s1=*p1;
		if(s1) {
			p2=p1;
			do {	//next s2
				p2++;
				s2=*p2;
				if(s2 && less(s2,s1)) {
					*p1=s2;
					*p2=s1;
					s1=s2;
				}
			} while(s2);
		}
		p1++;
	} while(s1);
}
